import warnings
from tmtccmd.fsfw.tmtc_printer import FsfwTmTcPrinter  # noqa: F401


warnings.warn(
    "This module is deprecated, use tmtccmd.fsfw.tmtc_printer",
    DeprecationWarning,
    stacklevel=2,
)
